//Library for converting and displaying nibbles, bytes, and strings

//It is assumed that the SEND_BUFFER_SIZE is defined already.
//If not, uncomment out the following default value:
#define SEND_BUFFER_SIZE 64

//Configuration Code required for use of Library  *******************************************
//The USART module must be set up to enable transmittion and it must be activated.
//The following sample code will set up the USART for full-duplex operation and set up
//the interrupts as low-priority. The USART handling code must be placed in the
//low-priority ISR in this case.
/*
    //Configure RS232 IO Pins
    TRISCbits.TRISC6 = 1;      //Configure Serial Port Pins
    TRISCbits.TRISC7 = 1;      //Configure Serial Port Pins

    //Configure Baud Rate Generator (Assuming 48 MHz Fosc)
    SPBRG = 77;                //Set Baud Rate to 9600Kbps
    SPBRGH = 0;                //Set Baud Rate to 9600Kbps
    BAUDCONbits.BRG16 = 0;     //Set USART to 8-bit (compatibility mode)
    BAUDCONbits.ABDEN = 0;     //Disable Baud rate detection
    TXSTAbits.BRGH = 0;        //Set Baud Rate Generator to Low Speed

    //Addittional Serial Port Configuration
    TXSTAbits.SYNC = 0;        //Asynchronus Mode
    RCSTAbits.SPEN = 1;        //Enable Serial Port
    TXSTAbits.TXEN = 1;        //Enable Transmittion
    RCSTAbits.CREN = 1;        //Enable Receiver

    //Configure Interrupts
    RCONbits.IPEN = 1;         //Enable Prioritized Interrupts
    IPR1bits.TXIP = 0;         //Set Transmit Interrupts to low priority
    IPR1bits.RCIP = 0;         //Set Receive Interrupts to low priority
    PIE1bits.TXIE = 1;         //Enable Transmit Interrupts
    PIE1bits.RCIE = 1;         //Enable Receive Interrupts

    INTCONbits.GIE = 1;        //Enable High Priority Interupts
    INTCONbits.PEIE = 1;       //Enable Low Priority Interrupts
*/
//Interrupts for transmitting over USART must be enabled and
//the relavent ISR must contain the following:
/*
    if (PIR1bits.TXIF == 1) {                       //Transmit Buffer is empty
        if (!(start == end)) {
            //There is something in the buffer to send
            TXREG = sendBuffer[start++];
            if (start == SEND_BUFFER_SIZE) start=0; //Wrap pointer to start of circular buffer
        }
    else PIE1bits.TXIE = 0; //Disable USART Transmit Interrupts
    }
*/
//End Configuration Code required for use of Library  ***************************************

//Global USART Buffer
volatile char sendBuffer[SEND_BUFFER_SIZE];         //Circular Buffer for transmittion
volatile char start=0;
volatile char end=0;

//Forward Declarations
void printStringRAM(char *str);                     //Print a null-terminated string stored in RAM
void printStringROM(far char *strPtr);              //Print a null-terminated string stored in ROM
void dumpStringRAM(char *strPtr);                   //Print a null-terminated string stored in RAM- wait for every character to send
void dumpStringROM(far char *strPtr);               //Print a null-terminated string stored in ROM- wait for every character to send (for long strings)
void printHex(char buffer[], char numberOfBytes);   //Print Multiple bytes of array in hex (spaces between bytes)
void printByteBinary(char input);                   //Print a single byte in Binary
void printDigit(unsigned char input);               //If a char has value 0-9, print the digit
void printNumber(unsigned char input);              //Print unsigned number stored in char
void echo(char input);                              //Print single character plus CR, LF and Prompt
char nibbleToChar(char nibble);                     //Returns Hex Character representing lower 4 bits of "nibble"
void byteToHexString(char currentByte, char *hexString);    //Dump two characters to target array for displaying byte in hex
void loadStringFromROM(far char *pointer, char *buffer);    //Copy multiple bytes from Program Memory to Data Memory
                                                            //This is good for storing lots of strings in ROM and pulling
                                                            //them into Data memory only when they are needed.
void flushBuffer(void);                             //Send all contents of outgoing buffer before continuing

//Global String Constants (RAM)
char CRandLF[3] = {13, 10, '\0'};
char prompt[3] = {'-', '>', '\0'};
char space[2] = {' ', '\0'};
char equals[2] = {'=', '\0'};

void echo(char input) {
    //write input, CR, LF, ->
    sendBuffer[end++] = input;
    if (end == SEND_BUFFER_SIZE) end=0;
    sendBuffer[end++] = 13;
    if (end == SEND_BUFFER_SIZE) end=0;
    sendBuffer[end++] = 10;
    if (end == SEND_BUFFER_SIZE) end=0;
    sendBuffer[end++] = '-';
    if (end == SEND_BUFFER_SIZE) end=0;
    sendBuffer[end++] = '>';
    if (end == SEND_BUFFER_SIZE) end=0;
    PIE1bits.TXIE = 1; //Enable USART Transmittion
}

void printNumber(unsigned char input) {
    //Print a number stored in a char
    char digit1, digit2, digit3;
    digit1 = input/100;
    if (digit1 > 0) printDigit(digit1);
    digit2 = (input - digit1*100)/10;
    if (digit1 > 0) printDigit(digit2);
    else if (digit2 > 0) printDigit(digit2);
    digit3 = input - digit1*100 - digit2*10;
    printDigit(digit3);
}

void printDigit(unsigned char input) {
    //Print a single digit- stored in a char
    //Print X if unrecognized
    char letter;
    switch (input) {
        case 0: letter = 48; break;
        case 1: letter = 49; break;
        case 2: letter = 50; break;
        case 3: letter = 51; break;
        case 4: letter = 52; break;
        case 5: letter = 53; break;
        case 6: letter = 54; break;
        case 7: letter = 55; break;
        case 8: letter = 56; break;
        case 9: letter = 57; break;
        default: letter = 88; break;
    }
    sendBuffer[end++] = letter;
    if (end == SEND_BUFFER_SIZE) end=0;
    PIE1bits.TXIE = 1; //Enable USART Transmittion
}

void printStringRAM(char *str) {
    //Dump Null-terminated string in RAM to USART transmit buffer
    char counter=0;
    char temp = str[counter++];
    while (!(temp == '\0')) {
        //The following line may cause firmware freeze if function is called from a thread of
        //higher priority than the USART handling ISRs- i.e. this line says to wait for USART to
        //catch up and clear buffer space for new string- if this catching up never occurs,
        //The loop never exits- hence, it has been commented out. The effect of this is that
        //if the buffer overflows, output over USART will be lost and strings may appear to be
        //scrambled. If this happens, increase USART buffer size.

        //while ((start == end+1) || ((start == 0) && (end == SEND_BUFFER_SIZE-1)));
        sendBuffer[end++] = temp;
        if (end == SEND_BUFFER_SIZE) end=0;
        temp = str[counter++];
    }
    PIE1bits.TXIE = 1; //Enable USART Transmittion
}

void printStringROM(far char *strPtr) {
    //Dump Null-terminated string in ROM to USART transmit buffer
    char index = 0;
    do {
        TBLPTR = strPtr[index++];
        sendBuffer[end++] = TABLAT;
        if (end == SEND_BUFFER_SIZE) end=0; //Wrap to front of circular buffer
    } while (TABLAT != '\0');
    PIE1bits.TXIE = 1; //Enable USART Transmittion
}

void dumpStringRAM(char *strPtr) {
     //Print a null-terminated string stored in RAM- wait for each character to send before sending next
     unsigned short long index = 0;
     char temp;
     PIE1bits.TXIE = 0;         //Disable Transmit Interrupts
     PIE1bits.RCIE = 0;         //Disable Receive Interrupts
     temp = strPtr[index++];
     while (temp != '\0') {
          TXREG = temp;
          Delay1TCY();
          while(PIR1bits.TXIF == 0);
          temp = strPtr[index++];
     }
     PIE1bits.TXIE = 1;         //Re-Enable Transmit Interrupts
     PIE1bits.RCIE = 1;         //Re-Enable Receive Interrupts
}

void dumpStringROM(far char *strPtr) {
     //Print a null-terminated string stored in ROM- wait for each character to send before sending next (for long strings)
     unsigned short long index = 0;
     char temp;
     PIE1bits.TXIE = 0;         //Disable Transmit Interrupts
     PIE1bits.RCIE = 0;         //Disable Receive Interrupts
     temp = strPtr[index++];
     while (temp != '\0') {
          TXREG = temp;
          Delay1TCY();
          while(PIR1bits.TXIF == 0);
          temp = strPtr[index++];
     }
     PIE1bits.TXIE = 1;         //Re-Enable Transmit Interrupts
     PIE1bits.RCIE = 1;         //Re-Enable Receive Interrupts
}

void printByteBinary(char input) {
    char buffer[9];
    if (input & 0b10000000) buffer[0]='1'; else buffer[0]='0';
    if (input & 0b01000000) buffer[1]='1'; else buffer[1]='0';
    if (input & 0b00100000) buffer[2]='1'; else buffer[2]='0';
    if (input & 0b00010000) buffer[3]='1'; else buffer[3]='0';
    if (input & 0b00001000) buffer[4]='1'; else buffer[4]='0';
    if (input & 0b00000100) buffer[5]='1'; else buffer[5]='0';
    if (input & 0b00000010) buffer[6]='1'; else buffer[6]='0';
    if (input & 0b00000001) buffer[7]='1'; else buffer[7]='0';
    buffer[8] = '\0';
    printStringRAM(buffer);
}

void printHex(char buffer[], char numberOfBytes) {
    char index;
    char tempbuf[3];
    tempbuf[2] = '\0';
    for (index=0; index<numberOfBytes; index++) {
        byteToHexString(buffer[index],tempbuf);
        printStringRAM(tempbuf);
        printStringRAM(space);
    }
}

void byteToHexString(char currentByte, char *hexString) {
    hexString[0] = nibbleToChar(currentByte >> 4);
    hexString[1] = nibbleToChar(currentByte & 0x0F);
}

char nibbleToChar(char nibble) {
    switch(nibble) {
        case 0: return('0');
        case 1: return('1');
        case 2: return('2');
        case 3: return('3');
        case 4: return('4');
        case 5: return('5');
        case 6: return('6');
        case 7: return('7');
        case 8: return('8');
        case 9: return('9');
        case 10: return('A');
        case 11: return('B');
        case 12: return('C');
        case 13: return('D');
        case 14: return('E');
        case 15: return('F');
        default: return('*');
    }
}

void loadStringFromROM(far char *pointer, char *buffer) {
    char index = 0;
    do {
        TBLPTR = pointer[index];
        buffer[index++] = TABLAT;
    } while (TABLAT != '\0');
}

void flushBuffer(void) {
     PIE1bits.TXIE = 0;         //Disable Transmit Interrupts
     PIE1bits.RCIE = 0;         //Disable Receive Interrupts
     while (!(start == end)) {
            //There is something in the buffer to send
            TXREG = sendBuffer[start++];
            if (start == SEND_BUFFER_SIZE) start=0;    //Wrap pointer to start of circular buffer
            while(PIR1bits.TXIF == 0);                 //Wait for byte to send
     }
     PIE1bits.TXIE = 1;         //Re-Enable Transmit Interrupts
     PIE1bits.RCIE = 1;         //Re-Enable Receive Interrupts
}
