//System Includes
#include <p18f13k50.h>
#include <delays.h>
#include "L3GD20TR_MemoryMap.h"
//#include "display.c"

//Definitions - Aliases
#define SS PORTCbits.RC6
#define GYRO_INT PORTCbits.RC2

//Set Configuration Bits
#pragma config CPUDIV = NOCLKDIV    //No Clock Division
#pragma config FOSC = HS            //External Crystal - High Speed Mode
#pragma config PLLEN = ON           //x4 PLL Enabled
#pragma config FCMEN = OFF          //Fail-Safe Clock Monitor Disabled
#pragma config PWRTEN = ON          //Power-up Timer Enabled
#pragma config BOREN = OFF          //Brown-Out Reset Disabled
#pragma config WDTEN = OFF          //Watchdog Timer Disabled
#pragma config HFOFST = OFF         //Wait for internal clock to stabalize before switching to HS mode
#pragma config MCLRE = ON           //MCLR pin is enabled
#pragma config STVREN = ON          //Reset on stack overflow
#pragma config LVP = OFF            //Single-Supply ICSP disabled
#pragma config CPD = OFF            //Disable EEPORM read protection
#pragma config WRTD = OFF           //Disable EEPORM write protection

//Output packet format is:    SYNC, HWTime, Gyro_X, Gyro_Y, Gyro_Z, Multiplier
//Field sizes are (in bytes):   4      4       2      2        2        8
//Field Types:                uint   uint     int     int     int     double
//All multi-byte fields are sent least-significant byte first. SYNC is 3325978160U
//HWTime is measured in ms. Multiply a Gyro reading by Multiplier to get the value
//in rad/s.

//Input interface is very primitive. Only one packet type is recognized:
//                            SYNC, Divider
//Field sizes are (in bytes):   4      1
//Field Types:                uint    uint
//Output data is streamed at the max rate of 95 Hz over this divider.
//Data is sampled at 95 Hz. If the divider is >1, measurements between outputs are
//averaged. SYNC is 1274290081U

//Global Constants and Variables
#define UART_BUFFER_SIZE 64
unsigned char UART_TXBuffer[UART_BUFFER_SIZE];
unsigned char UARTBufferStart, UARTBufferTail;

unsigned char GyroOutputDivider;

//Gyro Measurement Buffer
int GyroX, GyroY, GyroZ;
unsigned long long HWTime;

//Measurement Accumulation buffers
unsigned char sampleCounter;
signed long long GyroXAcc, GyroYAcc, GyroZAcc;

//forward declarations
void init(void);
void grabGyroMeasurement(void);
void sendMeasurementToHost(void);
void processReceivedCharacter(unsigned char byte);
void printByteBinary(char input);

//Driver function declarations
unsigned char L3GD20TR_READ(unsigned char address);
void L3GD20TR_MULTIREAD(unsigned char address, unsigned char numBytes, unsigned char * targetBuffer);
void L3GD20TR_Write(unsigned char address, unsigned char value);

//Interrupt Service Routines
void interrupt high_priority high_ISR (void) {
    //This will be invoked by Timer2 PR2 Matches every .001 seconds
    //It will also be tripped on UART interrupts
    if (PIR1bits.TMR2IF == 1) {
        PIR1bits.TMR2IF = 0;    //Clear the timer flag
        HWTime++;
    }
        
    if (PIR1bits.TXIF == 1) {
        //Transmit Buffer is empty
        if (UARTBufferStart != UARTBufferTail) {
            //There is something in the buffer to send
            TXREG = UART_TXBuffer[UARTBufferStart++];
            if (UARTBufferStart == UART_BUFFER_SIZE) UARTBufferStart=0; //Wrap pointer to start of circular buffer
        }
        else PIE1bits.TXIE = 0; //Disable USART Transmit Interrupts
    }

    if (PIR1bits.RCIF == 1) {
        processReceivedCharacter(RCREG);
    }
}

void processReceivedCharacter(unsigned char byte) {
    //Input interface is very primitive. Only one packet type is recognized:
    //                            SYNC, Divider
    //Field sizes are (in bytes):   4      1
    //Field Types:                uint    uint
    
    static unsigned char RXBuffer[5];
    static unsigned char byteCount;
    static unsigned long long lastEventTimestamp;
    static unsigned long long SYNC = 1274290081U;

    if (HWTime - lastEventTimestamp > 200) {
        byteCount = 0U;
        RXBuffer[byteCount++] = byte;
        lastEventTimestamp = HWTime;
    }
    else {
        RXBuffer[byteCount++] = byte;
        lastEventTimestamp = HWTime;
        if (byteCount == 5) {
            //process packet in RX buffer
            if ((RXBuffer[0] == ((unsigned char) SYNC))         && (RXBuffer[1] == (unsigned char) (SYNC >> 8)) &&
                (RXBuffer[2] == ((unsigned char) (SYNC >> 16))) && (RXBuffer[3] == ((unsigned char) (SYNC >> 24)))) {

                if (byte > 0U) {
                    GyroOutputDivider = byte;
                    sampleCounter = 0U;
                    GyroXAcc = 0;
                    GyroYAcc = 0;
                    GyroZAcc = 0;
                }
                else {
                    Reset();
                }
            }
            byteCount = 0U; //Flush RX buffer
        }
    }
}

void grabGyroMeasurement(void) {
    unsigned char byteL, byteH;
    byteL = L3GD20TR_READ(OUT_X_L);
    byteH = L3GD20TR_READ(OUT_X_H);
    GyroX = (int)(byteL + (byteH << 8));
    byteL = L3GD20TR_READ(OUT_Y_L);
    byteH = L3GD20TR_READ(OUT_Y_H);
    GyroY = (int)(byteL + (byteH << 8));
    byteL = L3GD20TR_READ(OUT_Z_L);
    byteH = L3GD20TR_READ(OUT_Z_H);
    GyroZ = (int)(byteL + (byteH << 8));
}

//********************   Driver for L3GD20TR   ********************
//*****************************************************************
//This driver is interruptable but non re-entrant. Only use from a
//single thread of execution or a single priority level.

unsigned char L3GD20TR_READ(unsigned char address) {
    //This function reads a single byte from the L3GD20TR
    unsigned char controlByte, temp, value;
    SSPCON1bits.WCOL = 0;
    controlByte = (address & 0b00111111) + 0b10000000;
    Delay1TCY();
    SS = 0;
    Delay1TCY();
    SSPBUF = controlByte;
    while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
    temp = SSPBUF;
    SSPBUF = 0U;
    while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
    value = SSPBUF;
    Delay1TCY();
    SS = 1;
    return(value);
}

void L3GD20TR_MULTIREAD(unsigned char address, unsigned char numBytes, unsigned char * targetBuffer) {
    //This function reads a single byte from the L3GD20TR
    unsigned char controlByte, index, temp;
    SSPCON1bits.WCOL = 0;

    if (numBytes == 0U) return; //Skip process if not requesting any data
    
    controlByte = (address & 0b00111111) + 0b11000000;
    Delay1TCY();
    SS = 0;
    Delay1TCY();
    SSPBUF = controlByte;
    while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
    temp = SSPBUF;
    for (index=0U; index<numBytes; index++) {
        SSPBUF = 0U;
        while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
        targetBuffer[index] = SSPBUF;
    }
    
    Delay1TCY();
    SS = 1;
}

void L3GD20TR_Write(unsigned char address, unsigned char value) {
    //This function writes a single byte to the L3GD20TR
    unsigned char byte, temp;
    SSPCON1bits.WCOL = 0;
    Delay1TCY();
    SS = 0;
    Delay1TCY();
    byte = address & 0b00111111;
    SSPBUF = byte;
    while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
    temp = SSPBUF;
    SSPBUF = value;
    while (SSPSTATbits.BF == 0);    //Wait for MSSP to finish transaction
    temp = SSPBUF;
    Delay1TCY();
    SS = 1;
}
//*****************************************************************
//*****************   End of Driver for L3GD20TR   ****************


void init(void){
    //*******     Configure MSSP for SPI mastering    *******
    //SPI subsystem will conform to standard SPI (1,1) mode. That is:
    //Base value of clock will be 1.
    //Data will be captured on rising edge of clock pulse and propogated on falling edge of pulse.
    PORTCbits.RC6 = 1;         //Set Data Latch for RC6 High (CS for Slave Device)
    TRISCbits.TRISC6 = 0;      //Set RC6 to be an output (CS for Slave Device)
    TRISCbits.TRISC7 = 0;      //Set SDO pin to output.
    TRISBbits.TRISB6 = 0;      //Set SCK pin to output.
    TRISBbits.TRISB4 = 1;      //Set SDI pin to input.
    SSPSTATbits.SMP = 0;       //Sample input on Rising edge of clock.
    SSPSTATbits.CKE = 0;       //Transmit occurs on falling edge of clock - should be sampled on rising edge
    SSPCON1bits.CKP = 1;       //Set Clock Idle State to 1 (Idle high)
    
    SSPCON1bits.SSPM = 0b0001; //Set to SPI Master Mode with clock set to FOSC/16 = 3 MHz

    SSPCON1bits.SSPEN = 1;     //Enable Subsystem
    //******   End Configure MSSP for SPI mastering    ******

    //Disable all Analog-to-Digital Functions
    ANSEL = 0U;
    ANSELH = 0U;

    //Set Up GYRO INT Pin
    TRISCbits.TRISC2 = 1;   //Set GYRO_INT to Input

    //Configure Timer2 for interrupts every 1 ms
    T2CONbits.T2OUTPS = 0b1001; //10:1 Postscaler
    T2CONbits.T2CKPS  = 0b10;   //16:1 Prescaler
    PR2 = 74U;                  //Reset every 75 increments (divide-by-75)
    T2CONbits.TMR2ON = 1;       //Enable Timer2
    
    //Configure USART for communication with MCP2200. Baud rate is 115.2 Kbaud
    TRISBbits.TRISB7 = 0;  //Set TX pin to output
    TRISBbits.TRISB5 = 1;  //Set RX pin to input
    TXSTAbits.TXEN = 1;    //Enable Transmitter
    RCSTAbits.CREN = 1;    //Enable Receiver
    TXSTAbits.SYNC = 0;    //Asynchronous Mode
    TXSTAbits.BRGH = 1;    //High speed baud-rate generation
    BAUDCONbits.BRG16 = 0; //8-bit baud-rate generation
    SPBRG = 25U;
    RCSTAbits.SPEN = 1;    //Enable Serial port

    //******* Begin L3GD20TR Initialization Proceedure *******
    //Initialize sensor with HP filter enabled
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG2, 0b00101001);  //Set HP filter cutoff frequency to 0.018 Hz
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG3, 0b00001000);  //Enable DRDY interrupt
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG4, 0b00010000);  //LSB@Lowest Address (Don't change Mem Map) and set full-scale range to 500 deg/s
    Delay100TCYx(1);
    //L3GD20TR_Write(CTRL_REG5, 0b00010001);  //Enable HP-filter. Grab data after LP filter 1 and HP filter
    L3GD20TR_Write(CTRL_REG5, 0b00000000);  //Disable HP-filter. Grab data after LP filter 1
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG1, 0b01011111);  //190 Hz 25 Hz cutoff.
    //L3GD20TR_Write(CTRL_REG1, 0b10111111);  //380 Hz 100 Hz cutoff.
    Delay100TCYx(1);
    L3GD20TR_READ(WHO_AM_I);                //First read after config is corrupted (not sure why) - flush it.
    //**********    End L3GD20TR Initialization    ***********

    //******* Begin L3GD20TR Initialization Proceedure *******
    //Initialize sensor with HP filter disabled
    /*Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG2, 0b00100100);  //Set HP filter cutoff frequency to 0.45 Hz
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG3, 0b00001000);  //Enable DRDY interrupt
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG4, 0b00010000);  //LSB@Lowest Address (Don't change Mem Map) and set full-scale range to 500 deg/s
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG5, 0b00000000);  //Disable HP-filter. Grab data after LP filter 1.
    Delay100TCYx(1);
    L3GD20TR_Write(CTRL_REG1, 0b00011111);
    Delay100TCYx(1);
    L3GD20TR_READ(WHO_AM_I);                //First read after config is corrupted (not sure why) - flush it.*/
    //**********    End L3GD20TR Initialization    ***********

    //************      Configure Interrupts     ************
    RCONbits.IPEN = 1;         //Enable Prioritized Interrupts
    
    //Configure Timer Interrupts
    IPR1bits.TMR2IP = 1;       //Set Timer2 interrupts to High-Priority
    PIR1bits.TMR2IF = 0;       //Clear TMR2 Flag
    PIE1bits.TMR2IE = 1;       //Enable Timer2 interrupts

    //Configure UART Interrupts
    IPR1bits.TXIP = 1;         //Set UART TX interrupts to high priority
    PIE1bits.TXIE = 1;         //Enable UART TX interrupts
    IPR1bits.RCIP = 1;         //Set UART RX interrupts to high priority
    PIE1bits.RCIE = 1;         //Enable UART RX interrupts

    //Enable Interrupts
    INTCONbits.GIEH = 1;       //Enable all unmasked High-Priority Interrupts
    INTCONbits.GIEL = 1;       //Enable all unmasked Low-Priority Interrupts
    //***********     End Configure Interrupts    ***********
}

void printByteBinary(char input) {
    char buffer[8];
    if (input & 0b10000000) buffer[0]='1'; else buffer[0]='0';
    if (input & 0b01000000) buffer[1]='1'; else buffer[1]='0';
    if (input & 0b00100000) buffer[2]='1'; else buffer[2]='0';
    if (input & 0b00010000) buffer[3]='1'; else buffer[3]='0';
    if (input & 0b00001000) buffer[4]='1'; else buffer[4]='0';
    if (input & 0b00000100) buffer[5]='1'; else buffer[5]='0';
    if (input & 0b00000010) buffer[6]='1'; else buffer[6]='0';
    if (input & 0b00000001) buffer[7]='1'; else buffer[7]='0';

    UART_TXBuffer[UARTBufferTail++] = '0';
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = 'b';
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[0];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[1];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[2];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[3];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[4];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[5];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[6];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = buffer[7];
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = '\n';
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = '\r';
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    PIE1bits.TXIE = 1;
}

void sendMeasurementToHost(void) {
    //Output packet format is:    SYNC, HWTime, Gyro_X, Gyro_Y, Gyro_Z, Multiplier
    //Field sizes are (in bytes):   4      4       2      2        2        8
    //Field Types:                uint   uint     int     int     int     double
    //All multi-byte fields are sent least-significant byte first. SYNC is 3325978160U
    //HWTime is measured in ms. Multiply a Gyro reading by Multiplier to get the value
    //in rad/s.
    static unsigned long long SYNC = 3325978160U;

    //Specify multiplier for 500 deg full-scale range: 0x3F31740A FA84AD8D
    static unsigned long long MultiplierLow  = 0xFA84AD8D;
    static unsigned long long MultiplierHigh = 0x3F31740A;

    unsigned long long latchedHWTime = HWTime;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) SYNC;
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (SYNC >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (SYNC >> 16);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (SYNC >> 24);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) latchedHWTime;
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (latchedHWTime >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (latchedHWTime >> 16);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (latchedHWTime >> 24);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) ((unsigned int) GyroXAcc);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (((unsigned int) GyroXAcc) >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) ((unsigned int) GyroYAcc);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (((unsigned int) GyroYAcc) >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) ((unsigned int) GyroZAcc);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (((unsigned int) GyroZAcc) >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) MultiplierLow;
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierLow >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierLow >> 16);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierLow >> 24);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    UART_TXBuffer[UARTBufferTail++] = (unsigned char) MultiplierHigh;
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierHigh >> 8);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierHigh >> 16);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;
    UART_TXBuffer[UARTBufferTail++] = (unsigned char) (MultiplierHigh >> 24);
    if (UARTBufferTail == UART_BUFFER_SIZE) UARTBufferTail=0;

    PIE1bits.TXIE = 1;
}

void main(void){
    unsigned char value;

    Delay10KTCYx(255U);
    
    init();

    //Set initial Gyro Output Divider to 1:95 to output data at 1 Hz
    GyroOutputDivider = 95U;

    //Read from the sensor
    value = L3GD20TR_READ(WHO_AM_I);
    printByteBinary(value);
    Delay100TCYx(100U);

    value = L3GD20TR_READ(CTRL_REG1);
    printByteBinary(value);
    Delay100TCYx(100U);

    value = L3GD20TR_READ(CTRL_REG2);
    printByteBinary(value);
    Delay100TCYx(100U);

    value = L3GD20TR_READ(CTRL_REG3);
    printByteBinary(value);
    Delay100TCYx(100U);

    value = L3GD20TR_READ(CTRL_REG4);
    printByteBinary(value);
    Delay100TCYx(100U);

    value = L3GD20TR_READ(CTRL_REG5);
    printByteBinary(value);
    Delay100TCYx(100U);

    //while(1);

    //Main Loop
    sampleCounter = 0U;
    GyroXAcc = 0;
    GyroYAcc = 0;
    GyroZAcc = 0;
    while(1) {
        if (GYRO_INT == 1) {
            grabGyroMeasurement();
            //Process new measurement
            GyroXAcc += (long long) GyroX;
            GyroYAcc += (long long) GyroY;
            GyroZAcc += (long long) GyroZ;
            sampleCounter++;
            if (sampleCounter == GyroOutputDivider) {
                GyroXAcc = GyroXAcc/GyroOutputDivider;
                GyroYAcc = GyroYAcc/GyroOutputDivider;
                GyroZAcc = GyroZAcc/GyroOutputDivider;

                sendMeasurementToHost();

                GyroXAcc = 0;
                GyroYAcc = 0;
                GyroZAcc = 0;
                sampleCounter = 0;
            }
        }
    }
}